import copy
import json
from typing import Any
from uuid import uuid4

from HTeaLeaf.Magic.jslib.JSCode import JSCode

from ..Html.Component import Component
from ..Html.Elements import div, script
from ..Server.Server import HttpRequest, Server, ServerEvent, Session


class SuperStore:
    _instance = None
    _initialized = False

    def __new__(cls, server=None):
        if cls._instance is None:
            cls._instance = super(SuperStore, cls).__new__(cls)
        return cls._instance


    def inject_stores(self, res_code, res_body, res_headers):
        if isinstance(res_body, Component):
            for store_id in self.stores:
                store = self.stores[store_id]
                res_body.append(script(f"const {store.do} = new Store(\"{store._id}\");"))

    def __init__(self, server: Server | None = None):
        if not self._initialized:
            self.stores: dict[str, Store | AuthStore] = {}
            self._initialized = True
            if server:
                server.add_path("/api/_store/{api_id}/*", self.process)
                server.registry_hook(ServerEvent.on_response, self.inject_stores)

            self._initialized = True

    def len(self):
        return len(self.stores)

    def add(self, id, store: "Store | AuthStore"):
        self.stores[id] = store

    def process(self,session: Session, req: HttpRequest, api_id):
        path = req.path.removeprefix(f"/api/_store/{api_id}/")

        store = self.stores.get(api_id)
        if store is None:
            return "Not found"

        if isinstance(store, AuthStore):
            store = store.auth(session)

        if store is None:
            return "404 Not Found", "Not found"


        if req.method == "GET":
            return json.dumps(store.read(path))
        elif req.method == "POST":
            data = req.json()
            text = req.text()
            if data is None:
                data = text
            return json.dumps(store.create(path, data))
        elif req.method == "DELETE":
            return json.dumps(store.delete(path))
        elif req.method == "PATCH":
            data = req.json()
            text = req.text()
            if data is None:
                data = text
            return json.dumps(store.update(path, data))
        else:
            return "404 Not Found", "Not found"



class Store:
    def __init__(self, default={}, subscribe=True, id = str(uuid4())):
        self._id = id
        self.data = copy.copy(default)
        self.do = JSCode(f"store_{self._id[:8]}")
        if subscribe:
            SuperStore().add(self._id, self)


    def __get_pointer__(self, path):
        pointer = self.data
        for item in path:
            # if not isinstance(pointer, Iterable):
            #     return None
            if type(pointer) is list:
                item = int(item)
                pointer = pointer[item]
            else:
                if item in pointer:
                    pointer = pointer[item]
                else:
                    return None
        return pointer

    def delete(self, path):
        path = path.split("/") if path != "" else []

        parent = self.__get_pointer__(path[:-1])
        if parent is None:
            return None
        item = path[-1]
        if type(parent) is list:
            del parent[int(item)]
            return True
        else:
            if item in parent:
                del parent[item]
                return True
            else:
                return False


    def update(self, path, data):
        path = path.split("/") if path != "" else []

        parent = self.__get_pointer__(path[:-1])
        item = path[-1]
        if parent is None:
            return None
        parent[item] = data

        return data


    def read(self, path: str) -> Any:
        path_list = path.split("/") if path != "" else []
        pointer = self.__get_pointer__(path_list)
        return pointer


    def react(self,path) -> Component:
        return div(self.read(path)).classes(f"{self._id}{id}_react")

    def create(self, path: str, data):
        path_list = path.split("/") if path != "" else []
        parent = self.__get_pointer__(path_list[:-1])
        if parent is None:
            return None
        item = path_list[-1]

        if item in parent:
            pointer = parent[item]
            if type(pointer) is dict:
                parent[item].update(data)
            elif type(pointer) is list:
                if not hasattr(data, "key"):
                    data["key"] = str(uuid4())
                pointer.append(data)
            else:
                return None

        else:
            parent[item] = data

        return parent


class AuthStore():
    def __init__(self, auth, default={}) -> None:
        self._id = str(uuid4())
        self.default = default
        self.data: dict[str, Store] = {}
        self.auth_func = auth
        self.do = JSCode(f"store_{self._id[:8]}")
        SuperStore().add(self._id, self)

    def auth(self, session: Session) -> Store:
        key = self.auth_func(session)
        if key not in self.data:
            self.data[key] = Store(default=copy.deepcopy(self.default),subscribe=False, id=self._id)
        return self.data[key]




from dataclasses import dataclass
from typing import Any

from .JSCode import JSCode


@dataclass
class document(JSCode):
    def __init__(self):
        super().__init__("document")

    getElementById: JSCode
    getElementByClassName: JSCode
    getElementsByTagName: JSCode
    querySelector: JSCode
    querySelectorAll: JSCode
    createElement: JSCode


def Not(code: JSCode):
    return JSCode(f"!{code}")

def Set(code: JSCode, other: Any):
    return JSCode(f"{code.raw} = {other}")

def Eq(code: JSCode, other: Any):
    return JSCode(f"{code.raw} == {other}")

def Ne(code: JSCode, other: Any):
    return JSCode(f"{code.raw} != {other}")

def Gt(code: JSCode, other: Any):
    return JSCode(f"{code.raw} > {other}")

def Lt(code: JSCode, other: Any):
    return JSCode(f"{code.raw} < {other}")



console = JSCode("console")
window = JSCode("window")
alert = JSCode("alert")



import inspect
import json
import textwrap
from types import FunctionType
from typing import Any

from .py2js import transpile


class JSCode:
    def __init__(self, raw: str):
        self.raw = raw

    def __str__(self):
        return self.raw

    def __repr__(self):
        return self.raw

    def __invert__(self):
        return JSCode(f"!{self.raw}")

    def __setattr__(self, name: str, value: Any, /) -> None:
        if name == "raw":
            object.__setattr__(self, name, value)
        else:
            pass

    def __getattr__(self, name: str):
        return JSCode(f"{self.raw}.{name}")

    def __setitem__(self, key, value):
        return JSCode(f"{self.raw}[{key}] = {value}")

    def __delitem__(self, key):
        return JSCode(f"delete {self.raw}[{key}]")

    def __add__(self, other):
        return JSCode(f"({self.raw} + {other})")

    def __sub__(self, other):
        return JSCode(f"({self.raw} - {other})")

    def __mul__(self, other):
        return JSCode(f"({self.raw} * {other})")

    def __truediv__(self, other):
        return JSCode(f"({self.raw} / {other})")

    def eq(self, other):
        return JSCode(f"({self.raw} == {other})")

    def ne(self, other):
        return JSCode(f"({self.raw} != {other})")

    def gt(self, other):
        return JSCode(f"({self.raw} > {other})")

    def lt(self, other):
        return JSCode(f"({self.raw} < {other})")


    def call(self, *args):
        from ..Store import AuthStore, Store
        parsed = []
        for arg in args:
            if isinstance(arg, JSCode):
                arg = arg.raw
            elif isinstance(arg, str):
                arg = json.dumps(arg)
            elif isinstance(arg, bool):
                arg = str(arg).lower()
            elif isinstance(arg, Store) or isinstance(arg, AuthStore):
                arg = str(arg.do())
            else:
                arg = str(arg)

            parsed.append(arg)

        payload = ",".join(parsed)
        return JSCode(f"{self.raw}({payload})")


    def __call__(self, *args: Any):
        return self.call(*args)



class JSFunction():
    def __init__(self,name: str, raw: str):
        super().__init__()
        self.name = name
        self.raw = raw

    def __str__(self):
        return self.raw

    def __call__(self, *args):
        for arg in args:
            if isinstance(arg, JSCode):
                arg = arg.raw
        return JSCode(f"{self.name}")(*args)


def js(fn: FunctionType):
    src = textwrap.dedent(inspect.getsource(fn))
    lines = src.splitlines()
    lines = [l for l in lines if not l.strip().startswith("@js")]
    src = "\n".join(lines)
    return JSFunction(fn.__name__, transpile(src))



from types import FunctionType
from typing import Any, List, Union

from HTeaLeaf.Magic.jslib import JSCode, JSFunction, js

from .Component import Component, ComponentMeta


class html(Component, metaclass=ComponentMeta):
    pass

class head(Component, metaclass=ComponentMeta):
    pass

class header(Component, metaclass=ComponentMeta):
    pass

class link(Component, metaclass=ComponentMeta):
    def __init__(self, *childs):
        super().__init__("link", *childs)


class script(Component):
    def __init__(self, *childs: Union[str, List[Any], "Component", JSFunction] ,src=None):
        parsed_childs: Union[str, List[Any], "Component"] = []
        for child in childs:
            if type(child) is FunctionType:
                parsed_childs.append(js(child))
            elif type(child) is JSFunction:
                parsed_childs.append(child)
            else:
                parsed_childs.append(child)
        super().__init__("script", *parsed_childs)
        self.unsafe = True
        if src is not None:
            self.attr(src=src)
            self.children = [""]
        # else:



class style(Component, metaclass=ComponentMeta):
    pass


class body(Component, metaclass=ComponentMeta):
    pass


class h1(Component, metaclass=ComponentMeta):
    pass


class h2(Component, metaclass=ComponentMeta):
    pass


class h3(Component, metaclass=ComponentMeta):
    pass


class div(Component, metaclass=ComponentMeta):
    pass

    def row(self):
        self.attr(style="display: flex; flex-direction: row")
        return self

    def column(self):
        self.attr(style="display: flex; flex-direction: column")
        return self


class button(Component, metaclass=ComponentMeta):

    def reactive(self,path,component_id):
        """
        Makes the button reactive by linking it to a FetchComponent.

        :param path: The URL to fetch new data from when clicked.
        :param component: The FetchComponent to be updated.
        """

        js = f"""fetchAndUpdate('{path}','{{}}','{component_id}')"""
        self.attr(onclick=js)
        return self



class label(Component, metaclass=ComponentMeta):
    pass

class checkbox(Component):
    def __init__(self,checked = False, *childs):
        super().__init__("input", *childs)
        self.attr(type="checkbox")
        if checked:
            self.attr(checked="True")

class textInput(Component):
    def __init__(self, *childs):
        super().__init__("input", *childs)

class select(Component):
    def __init__(self, items: List[str]):
        super().__init__("select")
        for item in items:
            self.append(option(item))

class option(Component):
    def __init__(self, value):
        super().__init__("option", value)
        self.attr(value=value)


class submit(Component):
    def __init__(self, *childs):
        super().__init__("input", *childs)
        self.attr(type="submit")

class form(Component):
    def __init__(self, *childs):
        super().__init__("form", *childs)

    def action(self, action):
        self.attr(action=action)
        return self

    def method(self, method):
        self.attr(method=method)
        return self

def tl_if(condition: JSCode | str | bool, *childs):
    if isinstance(condition, str):
        condition = JSCode(condition)

    return div(*childs).attr(style=f"display: {condition};")



import hashlib
import json
from typing import Any, List, Union

from ..Magic.jslib.JSCode import JSCode


class Component:
    """
    Represents an HTML component with attributes, children, and optional inline styles.
    This class allows constructing HTML elements programmatically and managing CSS styles.
    """

    def __init__(self, name, *childs: Union[str, List[Any], "Component", "JSCode"]) -> None:
        """
        Initializes a new Component instance.

        :param name: The tag name of the HTML element.
        :param childs: Optional children elements, which can be strings, lists, or other Component instances.
        """

        self.styles: str | None = None
        self.name = name
        self.children: list[Component | str | list | JSCode] = list(childs)
        self.attributes: dict[str, str | None] = dict()
        self._id: str = self._generate_id()

    def _generate_id(self) -> str:
        """Genera un ID determinista basado en el contenido del componente."""
        content = {
            "name": self.name,
            "children": [
                child._id if isinstance(child, Component) else str(child)
                for child in self.children
            ]
        }
        raw = json.dumps(content, sort_keys=True)
        hash_str = hashlib.md5(raw.encode()).hexdigest()[:12]
        return f"tl-{hash_str}"

    def id(self, id: str):
        """
        Sets the ID of the component and adds it as an attribute.

        :param id: The ID to assign.
        :return: The component instance (for method chaining).
        """

        self._id = id
        return self.attr(id=id)

    def classes(self, classes):
        """
        Adds a CSS class attribute to the component.

        :param classes: CSS class names (space-separated).
        :return: The component instance (for method chaining).
        """

        self.attributes["class"] = classes
        return self

    def style(self, path: str | None = None, **attr):
        """
        Adds inline styles to the component.

        :param path: Optional path to an external CSS file.
        :param attr: CSS properties to apply (e.g., color="red", margin="10px").
        :return: The component instance (for method chaining).
        """


        self.styles = (self.styles or "") + f"#{self._id} {{\n"
        self.styles += "\n".join(
            f"  {k.replace('_', '-')}: {v};" for k, v in attr.items()
        )
        self.styles += "\n}\n"

        if path:
            with open(path, "r") as f:
                self.styles += f.read()
        return self

    def attr(self,*args, **attr):
        """
        Adds custom attributes to the component.

        :param attr: Dictionary of attribute names and values.
        :return: The component instance (for method chaining).
        """

        for arg in args:
            self.attributes[arg] = None

        for k in attr:
            # if type(attr[k]) is str:
            value = attr[k]
            # if isinstance(value, JSCode):
            #     value = f"{{{{{str(value)}}}}}"
            self.attributes[k] = value
            # elif type(attr[k]) is FunctionType:
            #     py_f = inspect.getsource(attr[k])
            #     self.attributes[k] = f"""() => pyodide.runPython(`{py_f}`)"""

        return self

    def append(self, child: Union[str, "Component", list]):
        """
        Appends a child element to the component.

        :param child: A Component, string, or list of elements.
        :return: The component instance (for method chaining).
        """

        self.children.append(child)
        return self

    def prepend(self, child: Union[str, "Component", list]):
        """
        Prepends a child element to the component.

        :param child: A Component, string, or list of elements.
        :return: The component instance (for method chaining).
        """

        self.children.insert(0, child)
        return self

    def __build_attr__(self) -> str:
        return " " + " ".join(
            f"{k}='{v}'" if v is not None else f"{k}" for k, v in self.attributes.items()
        )

    def __build_child__(self, children: list):
        html_parts = []
        css_parts = []
        for child in children:
            if isinstance(child, str):
                html_parts.append(f"{child}")
            elif isinstance(child, list):
                html, css = self.__build_child__(child)
                html_parts.append(html)
                css_parts.append(css)
            elif isinstance(child, Component):
                html, css = child.build()
                html_parts.append(html)
                css_parts.append(css)
            # elif isinstance(child, JSDO):
            #     print(f"JSCode: {child().raw}")
            #     html_parts.append(f"{{{{{child().raw}}}}}")
            elif isinstance(child, JSCode):
                # JSCode outside of an attribute should be a special tag {{jscode_name}}
                print(f"JSCode: {child.raw}")
                html_parts.append(f"{{{{{child.raw}}}}}")
            else:
                try:
                    html_parts.append(str(child))
                except Exception:
                    continue
        return "".join(html_parts), "".join(filter(None, css_parts))

    def build(self) -> tuple[str, str]:
        """
        Builds the component's HTML and CSS separately.

        :return: A tuple (HTML string, CSS string)
        """

        if self.styles is not None and "id" not in self.attributes:
            self.attr(id=self._id)
        if len(self.children) == 0:
            result = f"<{self.name}{self.__build_attr__()}/>\n"
        else:
            endln = "\n" if len(self.children) > 1 else ""
            result = f"<{self.name}{self.__build_attr__()}>{endln}"
            html, styles = self.__build_child__(self.children)
            result += html
            if self.styles is None:
                self.styles = styles
            else:
                self.styles += styles
            result += f"\t</{self.name}>\n"
        css: str = "" if self.styles is None else self.styles
        return result, css

    def render(self) -> str:
        """
        Builds and returns the full HTML including inline CSS inside a <style> tag.

        :return: A complete HTML string with embedded CSS.
        """

        if len(self.children) == 0:
            result = f"<{self.name}{self.__build_attr__()}/>\n"
        else:
            inner_result, css = self.__build_child__(self.children)
            if self.styles is None:
                self.styles = css
            else:
                self.styles += css
                self.styles += "\n"
            result = f"<{self.name}{self.__build_attr__()}>\n"
            if self.styles is not None:
                result += f"<style>{self.styles}</style>\n"
            result += inner_result
            result += f"</{self.name}>\n"

        return result


class ComponentMeta(type):
    def __new__(cls, name, bases, dct):
        if name not in ("Component", "ComponentMeta"):

            def init(self, *childs):
                super(self.__class__, self).__init__(name, *childs)

            dct["__init__"] = init
        return super().__new__(cls, name, bases, dct)




import inspect
import json
import os
import re
import typing
from enum import Enum
from typing import Callable
from uuid import uuid4

from ..Html.Component import Component
from .Http.HttpHeader import Headers
from .Http.HttpRequest import HttpRequest
from .Http.HttpResponse import HttpResponse, HttpStatus


def path_to_regex(path: str) -> str:
    """
    Converts a path with curly braces and optional wildcards into a regex pattern.

    Examples:
        "/users/{id}"      -> "^/users/(?P<id>[^/]+)$"
        "/test/{id}/*"     -> "^/test/(?P<id>[^/]+)(?:/.*)?$"
        "/foo/*/bar"       -> "^/foo/.*/bar$"

    Args:
        path (str): The path pattern with placeholders and/or wildcards.

    Returns:
        str: The equivalent regex pattern.
    """

    # Sustituimos {param} por grupos con nombre
    regex = re.sub(r"\{([^}]+)\}", r"(?P<\1>[^/]+)", path)

    # Sustituimos '*' por un patrón que capture el resto del path
    regex = regex.replace("*", ".*")

    # Si acaba en '.*' (wildcard final), permitimos que sea opcional
    if regex.endswith(".*"):
        regex = regex[:-2] + "(?:.*)?"

    return f"^{regex}$"


def extract_wildcards(path_regex: str, url: str) -> dict | None:
    """
    Extracts wildcard values from a given regex pattern and a matching URL.

    Args:
        path_regex (str): The compiled regex pattern.
        url (str): The incoming request URL.

    Returns:
        dict | None: A dictionary of extracted values or None if no match.
    """
    match = re.match(path_regex, url)
    if match:
        return match.groupdict()  # Devuelve un diccionario con los valores capturados
    return None


class Session(dict):
    """
    A session object that behaves like a dictionary but allows attribute-style access.
    """

    def has(self, attr):
        """Checks if a session attribute exists."""
        return self.get(attr) is not None

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(f"'Session' object has no attribute '{attr}'")

    def __setattr__(self, attr, value):
        self[attr] = value



def match_path(
    routes: dict[str, typing.Callable], path: str
) -> tuple[dict[str, str | object], typing.Callable] | None:
    """
    Matches a given path against registered route patterns.

    Args:
        routes (dict[str, Callable]): A dictionary mapping regex patterns to handlers.
        path (str): The request path.

    Returns:
        tuple[dict, Callable] | None: Matched parameters and handler function, or None if no match.
    """

    for regex, value in routes.items():
        match = re.match(regex, path)
        if match:
            return match.groupdict(), value
    return None


def return_helper():
    try:
        helper = open(os.path.dirname(__file__) + "/helper.js")
        return 200 , helper.read()
    except Exception as e:
        print(e)
        return "404 Not Found", "Not Found"


class ServerEvent(Enum):
    on_response = "on_response"
    on_request = "on_request"
    path_registered = "path_registered"
    new_session = "new_session"



class Server:
    """
    HTTP server handling routing and session management.
    """

    def __init__(self):
        self.routes = {}
        self.sessions: dict[str, Session] = {}
        self._hooks: dict[ServerEvent,list[Callable[..., None]]] = {event: [] for event in ServerEvent}
        self.add_path("/_engine/helper.js", return_helper)

    def registry_hook(self, event: ServerEvent, callback: Callable[..., None]):
        event_hooks = self._hooks.get(event)
        if event_hooks is None:
            raise Exception("event dont exist")
        event_hooks.append(callback)

    def __call_hook__(self, event: ServerEvent, *payload):
        events = self._hooks.get(event)
        if events is None:
            return
        for callback in events:
            callback(*payload)

    def __create_session__(self):  # TODO: move to Session class
        """Generates a unique session ID."""
        session_id = str(uuid4())
        self.sessions[session_id] = Session()
        self.__call_hook__(ServerEvent.new_session, session_id, self.sessions[session_id])
        return session_id

    def route(self, path):
        """Registers a function as a handler for a given route pattern."""

        def decorator(func):
            self.add_path(path, func)
            return func

        return decorator

    def add_path(self, path, func):
        """Manually adds a route-handler mapping."""

        path_regex = path_to_regex(path)
        self.__call_hook__(ServerEvent.path_registered, path, path_regex, func)
        self.routes[path_regex] = func

    def __handle_session__(self, cookies: dict):
        header_session_cookie = None
        if cookies.get("TeaLeaf-Session") is None:
            session_id = self.__create_session__()
            header_session_cookie = ("Set-Cookie", f"TeaLeaf-Session={session_id}")
        else:
            session_id = cookies["TeaLeaf-Session"]
            if self.sessions.get(session_id) is None:
                self.sessions[session_id] = Session()
                self.__call_hook__(ServerEvent.new_session, session_id, self.sessions[session_id])


        return self.sessions[session_id], header_session_cookie

    def __process_response__(self, response) -> HttpResponse:
        res_code = HttpStatus.Ok
        res_headers = Headers()
        if isinstance(response, tuple):
            if len(response) == 3:
                res_code, headers, res_body = response
                res_headers = Headers(headers)

            elif len(response) == 2:
                res_code, res_body = response
            else:
                res_body = response[0]
        else:
            res_body = response

        self.__call_hook__(ServerEvent.on_response, res_code, res_body, res_headers)
        content_type = "text/plain"
        if isinstance(res_body, Component):
            content_type = "text/html"
            res_body = res_body.render()
        elif type(res_body) is dict or type(res_body) is list:
            content_type = "application/json"
            res_body = json.dumps(res_body)

        res_headers.add("Content-Type", content_type)
        return HttpResponse(res_code, res_headers, res_body)


    def handle_request(self, request: HttpRequest) -> HttpResponse:
        handler_and_match = match_path(self.routes, request.path)
        self.__call_hook__(ServerEvent.on_request, request)
        if handler_and_match is None:
            return HttpResponse(HttpStatus.NotFound, [("Content-Type", "text/plain")], "Not Found")

        params, handler = handler_and_match
        params["req"] = request
        _cookies = request.headers.get("COOKIE") or ""
        cookies = {
            k.strip(): v.strip()
            for k, v in (c.split("=", 1) for c in _cookies.split(";") if "=" in c)
        }


        params["session"], session_header = self.__handle_session__(cookies)
        params["cookies"] = cookies
        sig = inspect.signature(handler)
        params = {k: v for k, v in params.items() if k in sig.parameters}
        response = handler(**params)
        response = self.__process_response__(response)
        if session_header is not None:
            response.headers.add(session_header[0], session_header[1])
        return response

    def serve(self, payload: str | Component):
        pass



from ..Html.Component import Component
from ..Html.Elements import script
from ..Server.Server import Server, ServerEvent


def enable_reactivity(server: Server):
    helper_script = script(src="_engine/helper.js")

    def event_handler(res_code, res_body, res_headers):
        if isinstance(res_body, Component):
            res_body.prepend(helper_script)

    server.registry_hook(ServerEvent.on_response, event_handler)




from typing import Iterator

from ..Html.Component import Component
from .Http.HttpHeader import Headers
from .Server import HttpRequest, Server


def to_list(headers: Headers) -> list[tuple[str, str]]:
    return [h for h in headers]

class WSGI(Server):
    def __init__(self):
        super().__init__()



    def wsgi_app(self, environ: dict[str,str], start_response) -> Iterator[bytes]:
        path = environ.get('PATH_INFO', '/')
        method = environ.get('REQUEST_METHOD', 'GET')
        headers = {}
        try:
            headers["content_length"] = int(environ.get("CONTENT_LENGTH", 0))  # Puede ser None o vacío
        except ValueError:
            headers["content_length"] = 0
        for k in environ:
            if k.startswith("HTTP_"):
                headers[k[5:]] = environ[k]
        body = environ.get('wsgi.input',"body")
        request = HttpRequest(method,path,headers=headers,body=body)
        response = self.handle_request(request)
        start_response(response.status.to_str(), to_list(response.headers))
        response_body: bytes = response.body.encode("utf-8") if isinstance(response.body, str) else response.body
        return iter([response_body])

    def serve(self, payload: str | Component):
        return super().serve(payload)



from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Iterable, Literal, Optional

from .Http.HttpHeader import Headers
from .Server import HttpRequest, Server


@dataclass
class Scope:
    type: Literal["http"] | str
    asgi: dict[str,str] #version and spec_version
    http_version: str
    method: str
    scheme: str
    path: str
    raw_path: bytes
    query_string: bytes
    root_path: str
    headers: Iterable[tuple[bytes, bytes]]
    client: tuple[str,int]
    server:  tuple[str, Optional[int]]
    state: Optional[dict[str,Any]]



# dict struct
# recive event:
#     type: str
#     body: bytes
#     more_body: bool # if True wait for all the body chunks


def response_start(status: int, headers: Iterable[tuple[bytes,bytes]], trailers: bool) -> dict[str,Any]:
    result = dict()
    result["type"] = "http.response.start"
    result["status"] = status
    result["headers"] = headers
    result["trailers"] = trailers
    return result



def response_body(body: bytes = b"", more_body=False) -> dict[str,Any]:
    result = dict()
    result["type"] = "http.response.body"
    result["body"] = body
    result["more_body"] = more_body
    return result



def to_list(headers: Headers):
    header_list = []
    for (k,v) in headers:
        header_list.append((k.encode("utf-8"), v.encode("utf-8")))
    return header_list


class ASGI(Server):
    def __init__(self):
        super().__init__()


    async def application(self,scope: dict,
                          receive: Callable[[], Awaitable[dict]],
                          send: Callable[[Any], Awaitable[None]]):
        body: bytes = bytes()
        more = True
        while more:
            event = await receive()
            more = event["more_body"]
            body += event["body"]
        headers = [(k.decode(), v.decode()) for k,v in scope["headers"]]
        request = HttpRequest(scope["method"],scope["path"],headers=headers,body=body)
        response = self.handle_request(request)
        body = response.body.encode("utf-8") if isinstance(response.body, str) else response.body

        await send(response_start(response.status.to_int(),iter(to_list(response.headers)),False))

        await send(response_body(body))



import os
import sys
from urllib.parse import parse_qs

from ..Html.Component import Component
from .Server import Server


class CGI(Server):
    def __init__(self) -> None:
        super().__init__()
        input_data = sys.stdin.read()

        query_string = os.environ.get("QUERY_STRING", "")
        method = os.environ.get("REQUEST_METHOD", "")
        cgi_vars: dict[str, object] = dict()

        if query_string:
            cgi_vars.update(parse_qs(query_string))

        if method == "POST" and input_data:
            cgi_vars.update(parse_qs(input_data))

        # Obtener algunas variables de entorno comunes
        cgi_vars["REQUEST_METHOD"] = method
        cgi_vars["CONTENT_TYPE"] = os.environ.get("CONTENT_TYPE", "")
        cgi_vars["QUERY_STRING"] = query_string
        cgi_vars["PATH_INFO"] = os.environ.get("PATH_INFO", "")
        cgi_vars["SCRIPT_NAME"] = os.environ.get("SCRIPT_NAME", "")
        cgi_vars["SCRIPT_FILENAME"] = os.environ.get("SCRIPT_FILENAME", "")

        self.method = method
        self.server_vars = cgi_vars

    def serve(self, payload: str | Component):
        headers: dict[str, str] = dict()
        for k in headers:
            print(f"{k}: {headers[k]}")
        print("\r\n")
        content = payload
        if isinstance(payload, Component):
            try:
                content = payload.render()
            except Exception:
                print("<h1>500 Internal Server Error</h1>")
        print(content)



import hashlib
import inspect
import json
import re
from typing import Any
from uuid import uuid4

from HTeaLeaf.Html.Elements import script

from .JSCode import JSCode


class JSDO:
    def __init__(self, object_name: str, arg: Any):
        frame = inspect.stack()[1]
        site = f"{frame.filename}:{frame.lineno}"

        raw = f"{object_name}:{site}:{json.dumps(arg, sort_keys=True)}"
        id = hashlib.md5(raw.encode()).hexdigest()[:12]
        self.obj_name = f"{object_name.lower()}_{id}"
        self.store_js = f"const {self.obj_name} = new {object_name}({json.dumps(arg)}); {self.obj_name}.id = '{self.obj_name}';"


    def __call__(self):
        return JSCode(self.obj_name)

    def __format_js__(self, func_name: str, *args) -> JSCode:

        def mark_js(obj):
            if isinstance(obj, JSCode):

                return f"__JS__:{obj.raw}"
            elif isinstance(obj, (list, tuple)):
                return [mark_js(o) for o in obj]
            elif isinstance(obj, dict):
                return {k: mark_js(v) for k, v in obj.items()}
            else:
                return obj

        marked = [mark_js(arg) for arg in args]
        payload = json.dumps(marked)
        payload = re.sub(r'"__JS__:(.*?)(?<!\\)"', lambda m: m.group(1).replace('\\"','"'), payload)
        base_js = f"""{self.obj_name}.{func_name}({payload[1:-1]})"""
        return JSCode(
            base_js
        )

    def new(self):
        return script(self.store_js)

    def get(self, *args):
        return self.__format_js__("get", *args)


    def set(self, *args):
        print("args:", args)
        return self.__format_js__("set", *args)

    def update(self, *args):
        return self.__format_js__("update", *args)

    def delete(self, *args):
        return self.__format_js__("delete", *args)



import json
import os
import uuid
from typing import Any

from ..Html.Component import Component
from ..Html.Elements import div, script


class FetchComponent(Component):
    """
    A component that fetches data from a given URL and updates its content dynamically.
    """

    def __init__(self, url, body: str | dict | None = None) -> None:
        """
        Initializes a FetchComponent that loads content asynchronously.

        :param url: The URL to fetch data from.
        :param body: Optional request body for POST requests.
        """

        self._reid = new_id()
        placeholder = div("Loading...").id(self._reid)
        super().__init__("div", placeholder)
        # Configuración de la petición
        config = {"method": "POST" if body is not None else "GET", "body": body }
        # Serializar la configuración en JSON para JS
        _js_file = os.path.dirname(__file__) + "/MagicComponent.js"
        url = json.dumps(url)
        config_js = json.dumps(config)
        _id = json.dumps(placeholder._id)
        js: str = f"fetchAndUpdate({url},{config_js},{placeholder._id})"

        self.append(script(js))

    def reid(self):
        """
        Returns the reactive ID of the component.
        """

        return self._reid

class rButton(Component):
    """
    A reactive button that triggers updates on FetchComponents.
    """

    def __init__(self, *childs):
        """
        Initializes an rButton.

        :param childs: The child elements (text or components) inside the button.
        """
        super().__init__("button", *childs)

    def reactive(self,path,component: FetchComponent):
        """
        Makes the button reactive by linking it to a FetchComponent.

        :param path: The URL to fetch new data from when clicked.
        :param component: The FetchComponent to be updated.
        """

        #config = {"method": "GET"}
        if not hasattr(component, "reid"):
            raise Exception("component is not reactive")
        id = component.reid()
        # Serializar la configuración en JSON para JS
        #config_js = json.dumps(config)
        js = f"""fetchAndUpdate('{path}','{{}}','{id}')"""
        self.attr(onclick=js)
        return self

    #def refresh(self, path)

class HydratedComponent(Component):
    """
    WIP🚧
    A component that supports server-side hydration for dynamic updates.
    """

    def __init__(self, *childs):
        """
        Initializes a HydratedComponent.

        :param childs: The child elements inside the component.
        """

        pass


class PoolComponent(Component):
    """
    WIP🚧
    A component that periodically fetches data from a URL and updates itself.
    """

    def __init__(self, child: Component, url: str, interval: int = 5000):
        """
        Initializes a PoolComponent.

        :param child: The component to be updated.
        :param url: The URL from which new data will be fetched.
        :param interval: The time interval (in milliseconds) for polling updates. Default is 5000ms.
        """
        super().__init__("div", child)
        js = f"""
        setInterval(() => fetchAndUpdate('{url}', '{{}}', '{child._id}'), {interval});
        """
        self.append(script(js))


def new_id():
    """
    Generates a new unique identifier for reactive components.
    """
    return "tlmg" + str(uuid.uuid4()).split("-")[0]


def to_js_type(data: Any) -> str:
    return json.dumps(data)



from enum import Enum

from .HttpHeader import Headers


class HttpResponse:
    status: "HttpStatus"
    headers: Headers
    body: str | bytes

    def __init__(self, status: "HttpStatus" | tuple[int,str] | int, headers: Headers | list[tuple[str,str]], body: str | bytes, / ):
        if isinstance(status, HttpStatus):
            self.status = status
        elif isinstance(status,tuple):
            self.status = HttpStatus.custom(status[0], status[1])
        elif isinstance(status, int):
            self.status = HttpStatus.from_int(status)
        else:
            raise Exception(f"invalid status: {status}")

        if isinstance(headers,Headers):
            self.headers = headers
        elif isinstance(headers,list):
            self.headers = Headers(headers)
        else:
            raise Exception("invalid headers")

        self.body = body



class HttpStatus(Enum):
    Ok = 200, "Ok"
    Created = 201, "Created"
    NoContent = 204, "No Content"

    Found = 302, "Found"

    BadRequest = 400, "Bad Request"
    Unauthorized = 401, "Unauthorized"
    Forbidden = 403, "Forbiden"
    NotFound = 404, "Not Found"
    Conflict = 409, "Conflict"

    InternalServerError = 500, "Internal Server Error"
    NotImplemented = 501, "Not Implemented"
    BadGateway = 502, "Bad Gateway"
    ServiceUnavailable = 503, "Service Unavailable"
    _Other = None, None # Placeholder


    @classmethod
    def custom(cls, code, message):
        instance = cls._Other
        instance._value_ = code, message
        return instance

    @classmethod
    def from_int(cls, code: int):
        for status in cls:
            if status.value[0] == code:
                return status
        raise Exception(f"Invalid status {code}")


    def to_int(self) -> int:
        if self.value[0] is None:
            raise Exception(f"Invalid status {self.value}")
        return self.value[0]


    def to_str(self) -> str:
        if self.value is None:
            raise Exception(f"Invalid status {self}")
        return f"{self.value[0]} {self.value[1]}"



BANNED_HEADERS = []

class Headers():
    def __init__(self, data: list[tuple[str,str]] | dict[str,str] = []):
        self._data: dict[str, str] = {}
        for k in data:
            if isinstance(k, str) and isinstance(data, dict):
                self._data[k] = data[k]
            else:
                self._data[k[0]] = k[1]

    def __iter__(self):
        for k in self._data:
            yield (k, self._data[k])


    def __getitem__(self, key: str):
        return self._data.get(key.lower())


    def __setitem__(self, name: str, value: str) -> None:
        if name.lower() in self._data:
            if name.lower() in BANNED_HEADERS:
                raise Exception("Invalid duplicated header")
            self._data[name] += "; " + value
        else:
            self._data[name] = value

    def add(self, key: str, value: str):
        self._data[key] = value

    def get(self, key: str):
        for k in self._data:
            if k.lower() == key.lower():
                return self._data.get(k)
        return None


    def __contains__(self, item):
        return item.lower() in self._data



import hashlib
import inspect
import json

from HTeaLeaf.Html.Elements import script

from ..Magic.jslib.JSCode import JSCode
from ..Magic.jslib.JSDO import JSDO

# class localState():
#     def __init__(self, init_state):
#         self.do = JSDO("LocalState",init_state)

#     def

#     def js(self):
#         return self.do.js()

#     def get(self):
#         return self.do.get()

#     def set(self, data):
#         return self.do.set(data)



def use_state(init_state):
    frame = inspect.stack()[1]
    site = f"{frame.filename}:{frame.lineno}"
    raw = f"{site}:{json.dumps(init_state, sort_keys=True)}"
    id = hashlib.md5(raw.encode()).hexdigest()[:12]
    name = f"localstate_{id}"

    def new():
        return script(f"const {name} = new LocalState({json.dumps(init_state, sort_keys=True)},\"{name}\");")


    return new, JSCode(name)




import io
import json
from typing import Any, Optional

from .HttpHeader import Headers

class HttpRequest:
    """
    Represents an HTTP request with attributes for method, path, headers, and body.
    """

    def __init__(
        self,
        method: str = "GET",
        path: str = "/",
        args: dict[str, str] = {},
        headers: list[tuple[str, str]] | dict[str,str] = [],
        body: str | bytes | io.BufferedReader | None = None,
    ):

        self.method: str = method
        self.path: str = path
        self.args: dict[str, str] = args
        self.headers: Headers = Headers(headers)
        self.body: str | bytes | io.BufferedReader | Any | None = body

    def text(self) -> str | None:
        return self.__body_to_text__()

    def __body_to_text__(self) -> str | None:
        content_length =  self.headers.get("content-length")
        if content_length is None:
            return None

        body_size = int(content_length or 0)
        if body_size == 0 or self.body is None:
            return None
        if isinstance(self.body, io.BufferedReader):
            if not self.body.closed and self.body.readable():
                return self.body.read(body_size).decode("utf-8")
            else:
                return None
        elif isinstance(self.body, bytes):
            return self.body.decode("utf-8")
        elif isinstance(self.body, str):
            return self.body
        elif hasattr(self.body, '__iter__'):
            result = b"".join([d for d in iter(self.body)])
            return result.decode("utf-8")
        else:
            raise ValueError(f"Invalid body type: {type(self.body)}")

    def form(self) -> dict[str, str] | None:
        """
        Parses form-encoded body data into a dictionary.

        Returns:
            dict[str, str] | None: A dictionary of form values or None if invalid.
        """

        body = self.__body_to_text__()
        if body is None:
            return None
        return dict(item.split("=", 1) for item in body.split("&") if "=" in item)

    def json(self) -> Optional[dict]:
        """
        Parses the request body as JSON.

        Returns:
            dict | None: A dictionary representation of the JSON body or None if invalid.
        """

        body = self.__body_to_text__()
        if body is None:
            return None
        try:
            return json.loads(body)
        except (json.JSONDecodeError, AttributeError):
            return None



import ast
from typing import Optional

# ---------------------------------------------------------------------------
# Scope
# ---------------------------------------------------------------------------

class Scope:
    """Tracks declared variables per lexical scope."""

    def __init__(self, parent: Optional["Scope"] = None):
        self.parent = parent
        self.declared: set[str] = set()

    def is_declared(self, name: str) -> bool:
        if name in self.declared:
            return True
        return self.parent.is_declared(name) if self.parent else False

    def declare(self, name: str) -> None:
        self.declared.add(name)


# ---------------------------------------------------------------------------
# Transpiler
# ---------------------------------------------------------------------------

class PythonToJS(ast.NodeVisitor):

    def __init__(self) -> None:
        super().__init__()
        self.scope = Scope()

    # -- Scope helpers -------------------------------------------------------

    def _push_scope(self) -> None:
        self.scope = Scope(parent=self.scope)

    def _pop_scope(self) -> None:
        assert self.scope.parent is not None, "Cannot pop root scope"
        self.scope = self.scope.parent

    def _declare(self, name: str) -> str:
        """Return 'let ' if first declaration in current scope, else ''."""
        if not self.scope.is_declared(name):
            self.scope.declare(name)
            return "let "
        return ""

    # -- Indentation helper --------------------------------------------------

    def _indent(self, code: str, level: int = 1) -> str:
        pad = "\t" * level
        return "\n".join(pad + line for line in code.splitlines())

    # -- Body helper ---------------------------------------------------------

    def _emit_body(self, stmts: list[ast.stmt]) -> str:
        lines = []
        for stmt in stmts:
            result = self.visit(stmt)
            if result is not None:
                lines.append(result)
            else:
                # Useful during development to catch unhandled nodes
                lines.append(f"/* TODO: {type(stmt).__name__} */")
        return "\n".join(lines)

    # -----------------------------------------------------------------------
    # Fallback
    # -----------------------------------------------------------------------

    def generic_visit(self, node: ast.AST):
        raise NotImplementedError(
            f"Unhandled node type: {type(node).__name__}\n{ast.dump(node)}"
        )

    # -----------------------------------------------------------------------
    # Statements
    # -----------------------------------------------------------------------

    def visit_Expr(self, node: ast.Expr) -> str:
        return self.visit(node.value) + ";"

    def visit_Assign(self, node: ast.Assign) -> str:
        target = node.targets[0]
        name = self.visit(target)
        value = self.visit(node.value)

        prefix = self._declare(name) if isinstance(target, ast.Name) else ""
        return f"{prefix}{name} = {value};"

    def visit_AugAssign(self, node: ast.AugAssign) -> str:
        target = self.visit(node.target)
        op = self.visit(node.op)
        value = self.visit(node.value)
        return f"{target} {op}= {value};"

    def visit_AnnAssign(self, node: ast.AnnAssign) -> str:
        # Type annotations are discarded; treat as plain assignment
        name = self.visit(node.target)
        if node.value is None:
            return f"let {name};"
        value = self.visit(node.value)
        prefix = self._declare(name) if isinstance(node.target, ast.Name) else ""
        return f"{prefix}{name} = {value};"

    def visit_Return(self, node: ast.Return) -> str:
        if node.value is None:
            return "return;"
        return f"return {self.visit(node.value)};"

    def visit_If(self, node: ast.If) -> str:
        test = self.visit(node.test)
        body = self._indent(self._emit_body(node.body))
        result = f"if ({test}) {{\n{body}\n}}"
        if node.orelse:
            # elif becomes a nested If in the AST
            if len(node.orelse) == 1 and isinstance(node.orelse[0], ast.If):
                result += f" else {self.visit(node.orelse[0])}"
            else:
                orelse = self._indent(self._emit_body(node.orelse))
                result += f" else {{\n{orelse}\n}}"
        return result

    def visit_While(self, node: ast.While) -> str:
        test = self.visit(node.test)
        body = self._indent(self._emit_body(node.body))
        return f"while ({test}) {{\n{body}\n}}"

    def visit_For(self, node: ast.For) -> str:
        target = self.visit(node.target)
        iter_ = self.visit(node.iter)
        self._push_scope()
        self.scope.declare(target)
        body = self._indent(self._emit_body(node.body))
        self._pop_scope()
        return f"for (const {target} of {iter_}) {{\n{body}\n}}"

    def visit_Break(self, node: ast.Break) -> str:
        return "break;"

    def visit_Continue(self, node: ast.Continue) -> str:
        return "continue;"

    def visit_FunctionDef(self, node: ast.FunctionDef) -> str:
        params = ", ".join(arg.arg for arg in node.args.args)
        self._push_scope()
        # Parameters count as declared in function scope
        for arg in node.args.args:
            self.scope.declare(arg.arg)
        body = self._indent(self._emit_body(node.body))
        self._pop_scope()
        return f"function {node.name}({params}) {{\n{body}\n}}\n"

    # Async functions — emit as async JS functions
    visit_AsyncFunctionDef = visit_FunctionDef  # type: ignore[assignment]

    # -----------------------------------------------------------------------
    # Expressions
    # -----------------------------------------------------------------------

    def visit_Name(self, node: ast.Name) -> str:
        print(f"node name: {node}")
        return node.id

    def visit_Attribute(self, node: ast.Attribute) -> str:
        return f"{self.visit(node.value)}.{node.attr}"

    def visit_Call(self, node: ast.Call) -> str:
        func = self.visit(node.func)
        args = [self.visit(arg) for arg in node.args]
        # keyword args as trailing object literal
        if node.keywords:
            kwargs = ", ".join(
                f"{kw.arg}: {self.visit(kw.value)}" for kw in node.keywords if kw.arg
            )
            args.append(f"{{{kwargs}}}")
        return f"{func}({', '.join(args)})"

    def visit_Constant(self, node: ast.Constant) -> str:
        return self._emit_constant(node, raw=False)

    def _emit_constant(self, node: ast.Constant, raw: bool) -> str:
        kind = type(node.value).__name__
        if kind == "str":
            return str(node.value) if raw else f'"{node.value}"'
        if kind == "bool":
            return str(node.value).lower()
        if kind == "NoneType":
            return "null"
        return str(node.value)  # int, float

    def visit_JoinedStr(self, node: ast.JoinedStr) -> str:
        parts = []
        for value in node.values:
            if isinstance(value, ast.Constant):
                parts.append(self._emit_constant(value, raw=True))
            elif isinstance(value, ast.FormattedValue):
                parts.append(self.visit(value))
            else:
                parts.append(self.visit(value))
        return f"`{''.join(parts)}`"

    def visit_FormattedValue(self, node: ast.FormattedValue) -> str:
        return f"${{{self.visit(node.value)}}}"

    # -- Collections ---------------------------------------------------------

    def visit_List(self, node: ast.List) -> str:
        elts = ", ".join(self.visit(e) for e in node.elts)
        return f"[{elts}]"

    def visit_Tuple(self, node: ast.Tuple) -> str:
        # Tuples become arrays in JS
        return self.visit_List(node)  # type: ignore[arg-type]

    def visit_Dict(self, node: ast.Dict) -> str:
        pairs = []
        for key, value in zip(node.keys, node.values):
            k = "null" if key is None else self.visit(key)
            pairs.append(f"{k}: {self.visit(value)}")
        return f"{{{', '.join(pairs)}}}"

    def visit_Subscript(self, node: ast.Subscript) -> str:
        return f"{self.visit(node.value)}[{self.visit(node.slice)}]"

    # -- Operators -----------------------------------------------------------

    def visit_BinOp(self, node: ast.BinOp) -> str:
        left = self.visit(node.left)
        right = self.visit(node.right)
        op = self.visit(node.op)
        if isinstance(node.op, ast.FloorDiv):
            return f"Math.floor({left} / {right})"
        if isinstance(node.op, ast.Pow):
            return f"Math.pow({left}, {right})"
        return f"{left} {op} {right}"

    def visit_UnaryOp(self, node: ast.UnaryOp) -> str:
        op = self.visit(node.op)
        operand = self.visit(node.operand)
        return f"{op}{operand}"

    def visit_BoolOp(self, node: ast.BoolOp) -> str:
        op = self.visit(node.op)
        return f" {op} ".join(self.visit(v) for v in node.values)

    def visit_Compare(self, node: ast.Compare) -> str:
        parts = [self.visit(node.left)]
        for op, comp in zip(node.ops, node.comparators):
            parts.append(self.visit(op))
            parts.append(self.visit(comp))
        return " ".join(parts)

    # Arithmetic ops
    def visit_Add(self, node: ast.Add) -> str:         return "+"
    def visit_Sub(self, node: ast.Sub) -> str:         return "-"
    def visit_Mult(self, node: ast.Mult) -> str:       return "*"
    def visit_Div(self, node: ast.Div) -> str:         return "/"
    def visit_Mod(self, node: ast.Mod) -> str:         return "%"
    def visit_FloorDiv(self, node: ast.FloorDiv) -> str: return "/"  # handled in BinOp
    def visit_Pow(self, node: ast.Pow) -> str:         return "**"   # handled in BinOp

    # Unary ops
    def visit_USub(self, node: ast.USub) -> str:       return "-"
    def visit_UAdd(self, node: ast.UAdd) -> str:       return "+"
    def visit_Not(self, node: ast.Not) -> str:         return "!"
    def visit_Invert(self, node: ast.Invert) -> str:   return "~"

    # Comparison ops
    def visit_Eq(self, node: ast.Eq) -> str:           return "==="
    def visit_NotEq(self, node: ast.NotEq) -> str:     return "!=="
    def visit_Lt(self, node: ast.Lt) -> str:           return "<"
    def visit_LtE(self, node: ast.LtE) -> str:        return "<="
    def visit_Gt(self, node: ast.Gt) -> str:           return ">"
    def visit_GtE(self, node: ast.GtE) -> str:        return ">="
    def visit_Is(self, node: ast.Is) -> str:           return "==="
    def visit_IsNot(self, node: ast.IsNot) -> str:    return "!=="

    # Boolean ops
    def visit_And(self, node: ast.And) -> str:         return "&&"
    def visit_Or(self, node: ast.Or) -> str:           return "||"


# ---------------------------------------------------------------------------
# Entry point
# ---------------------------------------------------------------------------

def transpile(source: str) -> str:
    tree = ast.parse(source)
    visitor = PythonToJS()
    parts = []
    for node in tree.body:
        result = visitor.visit(node)
        if result:
            parts.append(result)
    return "\n".join(parts)





# if __name__ == "__main__":
#     import sys
#     src = sys.stdin.read() if len(sys.argv) == 1 else open(sys.argv[1]).read()
#     print(transpile(src))
